//** Modified-Light version of PPolygon for TP**// 




#ifndef   	POLYGON_H_
# define   	POLYGON_H_

#include <V_3D.h>
#include <Segment.h>
#include <Triangle.h>
#include <vector>


/// Manipulation class for a Polygon
/** Calculation for embedeed polygon. The polygon has to be plane and oriented.
    The polygon has to be convex in order to be triangulated as a fan.
*/

class PPolygon
{
public:
  
  //*********************************************//
  // CONSTRUCTORS
  //*********************************************//
  
  /// Classical constructor
  PPolygon();
  /// Constructor for triangle
  PPolygon(const V_3D& x0,const V_3D& x1,const V_3D& x2);
  /// Constructor for quad
  PPolygon(const V_3D& x0,const V_3D& x1,const V_3D& x2,const V_3D& x3);
  /// Constructor for any Polygon
  PPolygon(const std::vector <V_3D>& v);
  /// Copy Constructor
  PPolygon(const PPolygon& poly);
  /// destructor
  ~PPolygon();



  //*********************************************//
  // Access
  //*********************************************//

  /// get the number of edges
  int size() const;

  /// get an edge as a segment
  Segment get_segment(int index) const;
  
  //*********************************************//
  // Manipulation
  //*********************************************//

  /// destroy the polygon
  int destroy();
  
  /// set a arbitrary polygon
  int set_polygon(const std::vector <V_3D>& v);
  /// set a triangle
  int set_polygon(const V_3D& x0,const V_3D& x1,const V_3D& x2);
  /// set a quad
  int set_polygon(const V_3D& x0,const V_3D& x1,const V_3D& x2,const V_3D& x3);
  
  /// add a vertex
  int add_vertex(const V_3D& v);
  /// add a vertex
  int add_vertex(double x0,double x1,double x2);
  /// add a set of vertex
  int add_vertex(const std::vector <V_3D>& v);
  
  /// triangulate the given polygon as a fan
  /** return as may triangle polygons as needed to triangulate this one */
  std::vector <PPolygon> triangulate() const;
  /// triangulate the given polygon as a fan
  /** return as may Triangle as needed to triangulate this one */
  std::vector <Triangle> get_triangulation() const;

  /// get normal to the polygon
  V_3D normal() const;

  /// if the polygon has only three vertices, convert it to triangle
  Triangle to_triangle() const;

  //*********************************************//
  // Informations
  //*********************************************//


  /// get shortest distance between point and the polygon
  double shortest_distance_to_point(const V_3D& x) const;
  
  /// get shortest position on the polygon between the polygon and the point
  /** return in type the type of point
       type = 0: interior point in the polygon
       type = 1: edge point
       type = 2: vertex point 
       
       A vertex is inside the polygon (in the plane) if it lies on one of its triangulation
  */
  V_3D closest_point(const V_3D& x,int *type) const;
  /// get shortest position on the polygon between the polygon and the  point(no type information)
  V_3D closest_point(const V_3D& x) const;


  /// get closest point of the polygon in the segment
  /** two possiblities:
      1. The segment crosses the polygon => the closest point is inside the polygon
      2. The segment does not cross it => the closes point is in the edge */
  V_3D closest_segment_point(const Segment& s) const;


  /// Intersection with the plane of equation <n,x-x0>=0
  /** The plane is defined by its normal n, and a point x0, so its equation is <n,x-x0>=0
      4 possibilities: _ no intersections (type 0) => return void vector
		       _ plane intersect on two distinct points (type 1) => return vector of size 2
		       _ plane intersect only on a vertex (type 2) => return vector of size 1
                       _ plane intersect along an edge of the Polygon (type 3 > return the two extreme vertices) => return a vector of size 2

   */
  std::vector <V_3D> plane_intersection(const V_3D& n,const V_3D& x0,int *type) const;
  ///  Intersection with the plane of equation <n,x-x0>=0
  /** Same as plane_intersection function, but with the detailed type intersection for each edge */
  std::vector <V_3D> plane_intersection(const V_3D& n,const V_3D& x0,int *type,std::vector <int> *type_edge) const;


  /// Return the intersection of the polygon and the half space defined by the oriented plane <n,x-x0>=0
  /** 3 Possibilies for type: 0 - unchanged polygon
                              1 - partly cutted polygon
			      2 - no more polygon at all
  */
  PPolygon half_space_intersection(const V_3D& n,const V_3D& x0,int *type);


  //*********************************************//
  // Subdivision
  //*********************************************//
  
  /// split the given polygon by linking consecutiv edges
  /** Return the set of new polygons 
      Creates only triangles excepted for the last one which links
      every mid_points
      Works well for triangles
   */
  std::vector <PPolygon> subdivide_mid_edge() const;
  /// split the given polygon by linking barycenter to the middle edges
  /** Return the set of new polygons 
      Creates only four vertices polygons
      Works well for quads
   */
  std::vector <PPolygon> subdivide_barycenter_mid_edge() const;

  //*********************************************//
  // Operators
  //*********************************************//

  /// get the vertex
  V_3D operator()(int index) const;
  /// get the vertex
  V_3D& operator()(int index);
  /// get the vertex
  V_3D operator[](int index) const;
  /// get the vertex
  V_3D& operator[](int index);

  /// operator =
  PPolygon& operator=(const PPolygon&);

  /// output
  friend ostream& operator << (ostream& flux, const PPolygon& _v);


private:
  std::vector <V_3D> x;

};


#endif 	    /* !POLYGON_H_ */
